""" Use the python logging interface for vidavis logs.
If casalog is available, then send log messages to it.
Otherwise, use the base python logging mechanisms.
"""
import logging

try:
    from casatasks import casalog
    CASALOG_AVAIL = True
except ImportError:
    CASALOG_AVAIL = False

def get_logger():
    """ Returns the singleton logger instance. """
    if _Logging.logger is not None:
        return _Logging.logger

    if CASALOG_AVAIL:
        _Logging.log_handler = CasalogHandler()

        curr_logger_class = logging.getLoggerClass()
        logging.setLoggerClass(CasalogLogger)
        _Logging.logger = logging.getLogger(__name__)
        logging.setLoggerClass(curr_logger_class)

        _Logging.logger.addHandler(_Logging.log_handler)
        _Logging.logger.propagate = False
    else:
        _Logging.logger = logging.getLogger(__name__)
    return _Logging.logger

# pylint: disable=too-few-public-methods
class _Logging:
    # singleton instance
    logger = None
    log_handler = None
# pylint: enable=too-few-public-methods

class CasalogHandler(logging.Handler):
    """ Logs to casalog, as appropriate """
    # map from python logging levels to casalog priorities
    level_to_priority = {
        "DEBUG": "DEBUG",
        "INFO": "INFO",
        "WARNING": "WARN",
        "ERROR": "SEVERE",
        "CRITICAL": "SEVERE"
    }

    def _level_to_levelname(self, level):
        if level >= 50:
            return "CRITICAL"
        if level >= 40:
            return "ERROR"
        if level >= 30:
            return "WARNING"
        if level >= 20:
            return "INFO"
        return "DEBUG"

    def setLevel(self, level):
        super().setLevel(level)
        if not isinstance(level, str):
            levelname = self._level_to_levelname(level)
        else:
            levelname = level
        casalog.filter(CasalogHandler.level_to_priority[levelname])

    def emit(self, record):
        # promote log level for casalog
        levelname = self._level_to_levelname(_Logging.logger.getEffectiveLevel())
        if levelname == "DEBUG":
            casalog.filter(CasalogHandler.level_to_priority["DEBUG"])

        # get log properties
        origin = (record.filename) if (record.filename is not None) else ("")
        casalog.origin(origin)
        priority = CasalogHandler.level_to_priority[record.levelname]
        self.format(record) # populates record.message

        # emit the log
        casalog.post(message=record.message, priority=priority, origin=record.funcName)

class CasalogLogger(logging.getLoggerClass()):
    ''' CASA implementation of logger '''
    def setLevel(self, level):
        super().setLevel(level)
        _Logging.log_handler.setLevel(level)



'''
Base class for ms plots
'''

import os
import time

from bokeh.plotting import show
import hvplot
import holoviews as hv
import numpy as np
import panel as pn

try:
    from toolviper.utils.logger import get_logger, setup_logger
    _HAVE_TOOLVIPER = True
except ImportError:
    _HAVE_TOOLVIPER = False

from vidavis.data.measurement_set._ms_data import MsData
from vidavis.toolbox import AppContext, get_logger

class MsPlot:

    ''' Base class for MS plots with common functionality '''

# pylint: disable=too-many-arguments, too-many-positional-arguments
    def __init__(self, ms=None, log_level="info", log_to_file=False, show_gui=False, app_name="MsPlot"):
        if not ms and not show_gui:
            raise RuntimeError("Must provide ms/zarr path if gui not shown.")

        # Set logger: use toolviper logger else casalog else python logger
        if _HAVE_TOOLVIPER:
            self._logger = setup_logger(app_name, log_to_term=True, log_to_file=log_to_file, log_file=app_name.lower(), log_level=log_level.upper())
        else:
            self._logger = get_logger()
            self._logger.setLevel(log_level.upper())

        # Save parameters; ms set below
        self._show_gui = show_gui
        self._app_name = app_name

        # Set up temp dir for output html files
        self._app_context = AppContext(app_name)

        if show_gui:
            # Enable "toast" notifications
            pn.config.notifications = True
            self._toast = None # for destroy() with new plot or new notification

        # Initialize plot inputs and params
        self._plot_inputs = {'selection': {}}
        self._plot_params = None

        # Initialize plots
        self._plot_init = False
        self._plots_locked = False
        self._plots = []

        # Set data (if ms)
        self._data = None
        self._ms_info = {}
        self._set_ms(ms)
# pylint: enable=too-many-arguments, too-many-positional-arguments

    def summary(self, data_group='base', columns=None):
        ''' Print ProcessingSet summary.
            Args:
                data_group (str): data group to use for summary.
                columns (None, str, list): type of metadata to list.
                    None:      Print all summary columns in ProcessingSet.
                    'by_msv4': Print formatted summary metadata by MSv4.
                    str, list: Print a subset of summary columns in ProcessingSet.
                        Options: 'name', 'intents', 'shape', 'polarization', 'scan_name', 'spw_name',
                                 'field_name', 'source_name', 'field_coords', 'start_frequency', 'end_frequency'
            Returns: list of unique values when single column is requested, else None
        '''
        if self._data:
            self._data.summary(data_group, columns)
        else:
            self._logger.error("Error: MS path has not been set")

    def data_groups(self):
        ''' Returns set of data groups from all ProcessingSet ms_xds. '''
        if self._data:
            return self._data.data_groups()
        self._logger.error("Error: MS path has not been set")
        return None

    def get_dimension_values(self, dimension):
        ''' Returns sorted list of unique dimension values in ProcessingSet (with previous selection applied, if any).
            Dimension options include 'time', 'baseline' (for visibility data), 'antenna' (for spectrum data), 'antenna1',
                'antenna2', 'frequency', 'polarization'.
        '''
        if self._data:
            return self._data.get_dimension_values(dimension)
        self._logger.error("Error: MS path has not been set")
        return None

    def plot_antennas(self, label_antennas=False):
        ''' Plot antenna positions.
                label_antennas (bool): label positions with antenna names.
        '''
        if self._data:
            self._data.plot_antennas(label_antennas)
        else:
            self._logger.error("Error: MS path has not been set")

    def plot_phase_centers(self, data_group='base', label_fields=False):
        ''' Plot the phase center locations of all fields in the Processing Set and highlight central field.
                data_group (str): data group to use for field and source xds.
                label_fields (bool): label all fields on the plot if True, else label central field only
        '''
        if self._data:
            self._data.plot_phase_centers(data_group, label_fields)
        else:
            self._logger.error("Error: MS path has not been set")

    def clear_plots(self):
        ''' Clear plot list '''
        while self._plots_locked:
            time.sleep(1)
        self._plots.clear()

    def clear_selection(self):
        ''' Clear data selection and restore original ProcessingSet '''
        if self._data:
            self._data.clear_selection()

        self._plot_inputs['selection'] = {}

    def show(self):
        ''' 
        Show interactive Bokeh plots in a browser. Plot tools include pan, zoom, hover, and save.
        '''
        if not self._plots:
            raise RuntimeError("No plots to show.  Run plot() to create plot.")

        # Do not delete plot list until rendered
        self._plots_locked = True

        # Single plot or combine plots into layout using subplots (rows, columns)
        # Not layout if subplots is single plot (default if None) or if only one plot saved
        subplots = self._plot_inputs['subplots']
        layout_plot, is_layout = self._layout_plots(subplots)

        # Render to bokeh figure
        if is_layout:
            # Show plots in columns
            plot = hv.render(layout_plot.cols(subplots[1]))
        else:
            # Show single plot
            plot = hv.render(layout_plot)

        self._plots_locked = False
        if self._plot_params:
            column = pn.Column()
            for param in self._plot_params:
                column.append(pn.pane.Str(param))
            tabs = pn.Tabs(('Plot', plot), ('Plot Inputs', column))
            tabs.show(title=self._app_name, threaded=True)
        else:
            show(plot)

    def save(self, filename='ms_plot.png', fmt='auto', width=900, height=600):
        '''
        Save plot to file with filename, format, and size.
        If iteration plots were created:
            If subplots is a grid, the layout plot will be saved to a single file.
            If subplots is a single plot, iteration plots will be saved individually,
                with a plot index appended to the filename: {filename}_{index}.{ext}.
        '''
        if not self._plots:
            raise RuntimeError("No plot to save.  Run plot() to create plot.")

        start_time = time.time()

        # Save single plot or combine plots into layout using subplots (rows, columns).
        # Not layout if subplots is single plot or if only one plot saved.
        subplots = self._plot_inputs['subplots']
        layout_plot, is_layout = self._layout_plots(subplots)

        if is_layout:
            # Save plots combined into one layout
            hvplot.save(layout_plot.cols(subplots[1]), filename=filename, fmt=fmt)
            self._logger.info("Saved plot to %s.", filename)
        else:
            # Save plots individually, with index appended if exprange='all' and multiple plots.
            if self._plot_inputs['iter_axis'] is None:
                hvplot.save(layout_plot.opts(width=width, height=height), filename=filename, fmt=fmt)
                self._logger.info("Saved plot to %s.", filename)
            else:
                name, ext = os.path.splitext(filename)
                iter_range = self._plot_inputs['iter_range'] # None or (start, end)
                plot_idx = 0 if iter_range is None else iter_range[0]

                for plot in self._plots:
                    exportname = f"{name}_{plot_idx}{ext}"
                    hvplot.save(plot.opts(width=width, height=height), filename=exportname, fmt=fmt)
                    self._logger.info("Saved plot to %s.", exportname)
                    plot_idx += 1

        self._logger.debug("Save elapsed time: %.2fs.", time.time() - start_time)

    def _layout_plots(self, subplots):
        subplots = (1, 1) if subplots is None else subplots
        num_plots = len(self._plots)
        num_layout_plots = min(num_plots, np.prod(subplots))

        if num_layout_plots == 1:
            return self._plots[0], False

        # Set plots in layout
        plot_count = 0
        layout = None
        for i in range(num_layout_plots):
            plot = self._plots[i]
            layout = plot if layout is None else layout + plot
            plot_count += 1

        is_layout = plot_count > 1
        return layout, is_layout

    def _set_ms(self, ms_path):
        ''' Set MsData and update ms info for input ms filepath (MSv2 or zarr), if set.
            Return whether ms changed (false if ms is None). '''
        self._ms_info['ms'] = ms_path
        ms_error = ""
        ms_changed = ms_path and (not self._data or not self._data.is_ms_path(ms_path))

        if ms_changed:
            try:
                # Set new MS data
                self._data = MsData(ms_path, self._logger)
                data_path = self._data.get_path()
                self._ms_info['ms'] = data_path
                root, ext = os.path.splitext(os.path.basename(data_path))
                while ext != '':
                    root, ext = os.path.splitext(root)
                self._ms_info['basename'] = root
                self._ms_info['data_dims'] = self._data.get_data_dimensions()
            except RuntimeError as e:
                ms_error = str(e)
                self._data = None

        if ms_error:
            self._notify(ms_error, 'error', 0)

        return ms_changed

    def _notify(self, message, level, duration=3000):
        ''' Log message. If show_gui, notify user with toast for duration in ms.
            Zero duration must be dismissed. '''
        if self._show_gui:
            pn.state.notifications.position = 'top-center'
            if self._toast:
                self._toast.destroy()

        if level == "info":
            self._logger.info(message)
            if self._show_gui:
                self._toast = pn.state.notifications.info(message, duration=duration)
        elif level == "error":
            self._logger.error(message)
            if self._show_gui:
                self._toast = pn.state.notifications.error(message, duration=duration)
        elif level == "success":
            self._logger.info(message)
            if self._show_gui:
                self._toast = pn.state.notifications.success(message, duration=duration)
        elif level == "warning":
            self._logger.warning(message)
            if self._show_gui:
                self._toast = pn.state.notifications.warning(message, duration=duration)

    def _set_plot_params(self, plot_params):
        ''' Set list of plot parameters as key=value string, for logging or browser display '''
        plot_inputs = plot_params.copy()
        for key in ['self', '__class__', 'data_dims']:
            # Remove keys from using function locals()
            try:
                del plot_inputs[key]
            except KeyError:
                pass
        self._plot_params = sorted([f"{key}={value}" for key, value in plot_inputs.items()])



'''
Create panel widgets for various functions
'''

import panel as pn

from vidavis.bokeh._palette import available_palettes
from vidavis.plot.ms_plot._ms_plot_constants import VIS_AXIS_OPTIONS, AGGREGATOR_OPTIONS, PS_SELECTION_OPTIONS, MS_SELECTION_OPTIONS, DEFAULT_UNFLAGGED_CMAP, DEFAULT_FLAGGED_CMAP

def file_selector(description, start_dir, callback):
    ''' Return a layout for file selection with input description and start directory.
        Includes a TextInput and a FileSelector, with a callback to set TextInput from FileSelector.
    '''
    filename = pn.widgets.TextInput(
        description=description,
        name="Filename",
        placeholder='Enter filename or use file browser below',
        sizing_mode='stretch_width',
    )

    file_select = pn.widgets.FileSelector(
        start_dir,
    )
    select_file = pn.bind(callback, file_select)

    fs_card = pn.Card(
        file_select,
        title='File browser',
        collapsed=True,
        collapsible=True,
        sizing_mode='stretch_width',
    )

    return pn.Column(
        pn.Row( # [0]
            filename,   # [0]
            select_file # [1]
        ),
        fs_card, # [1]
        width_policy='min',
    )

def style_selector(style_callback, color_range_callback):
    ''' Return a layout for style parameters.
        Currently supports colormaps, colorbar, and color limits.
    '''
    cmaps = available_palettes()

    cmap_selector = pn.widgets.Select(
        name="Unflagged data colormap",
        options=cmaps,
        value=DEFAULT_UNFLAGGED_CMAP,
        sizing_mode='scale_width',
    )

    flagged_cmap_selector = pn.widgets.Select(
        name="Flagged data colormap",
        options=cmaps,
        value=DEFAULT_FLAGGED_CMAP,
        sizing_mode='scale_width',
    )

    colorbar_checkbox = pn.widgets.Checkbox(
        name="Show colorbar",
        value=True,
    )

    flagged_colorbar_checkbox = pn.widgets.Checkbox(
        name="Show flagged colorbar",
        value=True,
    )

    select_style = pn.bind(style_callback, cmap_selector, flagged_cmap_selector, colorbar_checkbox, flagged_colorbar_checkbox)

    color_mode_selector = pn.widgets.RadioBoxGroup(
        options=["No color range", "Auto color range", "Manual color range"],
    )

    color_range_slider = pn.widgets.RangeSlider(
        name="Colorbar range",
    )

    select_color_range = pn.bind(color_range_callback, color_mode_selector, color_range_slider)

    return pn.Column(
        pn.Row( # [0]
            cmap_selector,         # [0]
            flagged_cmap_selector, # [1]
        ),
        pn.Row( # [1]
            colorbar_checkbox,         # [0]
            flagged_colorbar_checkbox, # [1]
        ),
        select_style, # [2]
        pn.Row( # [3]
            color_mode_selector, # [0]
            color_range_slider,  # [1]
        ),
        select_color_range, # [4]
        width_policy='min',
    )

def axis_selector(x_axis, y_axis, axis_options, include_vis, callback):
    ''' Return layout of selectors for x-axis, y-axis, and vis-axis '''
    x_options = axis_options if axis_options else [x_axis]
    x_selector = pn.widgets.Select(
        name="X Axis",
        options=x_options,
        value=x_axis,
        sizing_mode='scale_width',
    )

    y_options = axis_options if axis_options else [y_axis]
    y_selector = pn.widgets.Select(
        name="Y Axis",
        options=y_options,
        value=y_axis,
        sizing_mode='scale_width',
    )

    if include_vis:
        # for raster plot
        vis_label = pn.pane.LaTeX(
            r"Vis Axis",
        )
        vis_selector = pn.widgets.RadioBoxGroup(
            options=VIS_AXIS_OPTIONS,
            inline=True,
        )
        select_axes = pn.bind(callback, x_selector, y_selector, vis_selector)
        return pn.Column(
            pn.Row( # [0]
                x_selector, # [0]
                y_selector  # [1]
            ),
            vis_label,    # [1]
            vis_selector, # [2]
            select_axes,
            width_policy='min',
        )

    # for scatter plot
    select_axes = pn.bind(callback, x_selector, y_selector)
    return pn.Column(
        pn.Row( # [0]
            x_selector, # [0]
            y_selector  # [1]
        ),
        select_axes,
        width_policy='min',
    )


def aggregation_selector(axis_options, callback):
    ''' Return layout of selectors for aggregator and agg axes '''
    agg_selector = pn.widgets.Select(
        name="Aggregator",
        options=AGGREGATOR_OPTIONS,
        description="Select aggregation type",
        sizing_mode='scale_width',
    )

    agg_axis_selector = pn.widgets.MultiSelect(
        name="Agg axis or axes",
        options=axis_options,
        description="Select one or more axes to aggregate",
        sizing_mode='scale_width',
    )

    select_agg = pn.bind(callback, agg_selector, agg_axis_selector)

    return pn.Row(
        agg_selector,      # [0]
        agg_axis_selector, # [1]
        select_agg,
        width_policy='min',
    )

def iteration_selector(axis_options, axis_callback, iter_callback):
    ''' Return layout of selectors for iteration axis and player selector for value.
        Callback sets values in iter value player and iter value range selectors when iter_axis is selected.
    '''
    iter_options = ['None']
    iter_options.extend(axis_options)

    iter_axis_selector = pn.widgets.Select(
        name="Iteration axis",
        options=iter_options,
        description="Select axis over which to iterate",
        sizing_mode='scale_width',
    )

    update_iter_values = pn.bind(axis_callback, iter_axis_selector)

    iter_value_type = pn.widgets.RadioBoxGroup(
        options=['By Value', 'By Range'],
        inline=True,
    )

    # Single value
    iter_value_player = pn.widgets.DiscretePlayer(
        name="Iteration value",
        show_loop_controls=False,
        show_value=False,
        value_align='start',
        visible_buttons=['first', 'previous', 'next', 'last'],
        sizing_mode='scale_width',
    )

    # Range
    iter_range_start = pn.widgets.IntInput(
        name="Iteration start",
        start=0,
        value=0,
        description="Index of first value in iteration",
        sizing_mode='scale_width',
    )
    iter_range_end = pn.widgets.IntInput(
        name="Iteration end",
        start=0,
        value=0,
        description="Index of last value in iteration",
        sizing_mode='scale_width',
    )
    subplot_rows = pn.widgets.IntInput(
        name="Subplot rows",
        start=1,
        end=10,
        description="Number of rows to display iteration plots",
        sizing_mode='scale_width',
    )
    subplot_columns = pn.widgets.IntInput(
        name="Subplot columns",
        start=1,
        end=10,
        description="Number of columns to display iteration plots",
        sizing_mode='scale_width',
    )
    iter_range_widgets = pn.Column(
        pn.Row( # [0]
            iter_range_start, # [0]
            iter_range_end,   # [1]
        ),
        pn.Row( # [1]
            subplot_rows,    # [0]
            subplot_columns, # [1]
        ),
        pn.pane.LaTeX(
            r"Multiple subplots will be shown in new tab",
        ),
    )

    # Put iter value/range into accordion
    iter_value_selectors = pn.Accordion(
        ('Select Single Iteration Value', iter_value_player), # [0]
        ('Select Iteration Index Range', iter_range_widgets), # [1]
        toggle=True,
        sizing_mode='stretch_width',
    )

    select_iter = pn.bind(iter_callback, iter_axis_selector, iter_value_type, iter_value_player, iter_range_start, iter_range_end, subplot_rows, subplot_columns)

    return pn.Column(
        pn.Row( # [0]
            iter_axis_selector, # [0]
            iter_value_type,    # [1]
            update_iter_values,
        ),
        iter_value_selectors, # [1]
        select_iter,
        width_policy='min',
    )

def title_selector(callback):
    ''' Return a layout for title input using TextInput '''
    title_input = pn.widgets.TextInput(
        name="Title",
        placeholder="Enter title for plot ('ms' to use MS name)",
        sizing_mode='stretch_width',
    )
    select_title = pn.bind(callback, title_input)
    return pn.Row(
        title_input,
        select_title,
        width_policy='min',
    )

def selection_selector(ps_callback, ms_callback):
    ''' Return layout of selectors for ProcessingSet and MSv4 selection. '''
    # Create column of ps selectors; values added from PS summary when ms path is set.
    ps_selection = pn.Column(
        height_policy='min',
    )
    ps_selection.append(
        pn.widgets.TextInput(
            name="Query",
            placeholder='Enter query for summary columns',
            sizing_mode='stretch_width',
        )
    )
    _add_multi_choice(ps_selection, PS_SELECTION_OPTIONS) # 7 more options

    select_ps = pn.bind(ps_callback, ps_selection[0], ps_selection[1], ps_selection[2], ps_selection[3], ps_selection[4], ps_selection[5], ps_selection[6], ps_selection[7])
    ps_selection.append(select_ps)

    # Create column of ms selectors; values added from MS later.
    ms_selection = pn.Column()
    _add_multi_choice(ms_selection, MS_SELECTION_OPTIONS) # 7 options
    select_ms = pn.bind(ms_callback, ms_selection[0], ms_selection[1], ms_selection[2], ms_selection[3], ms_selection[4], ms_selection[5], ms_selection[6])
    ms_selection.append(select_ms)

    selection_cards = pn.Accordion(
        ("Select ProcessingSet", ps_selection),  # [0] pn.Column
        ("Select MeasurementSet", ms_selection), # [1] pn.Column
        toggle=True,
        sizing_mode='stretch_width',
    )

    return pn.Column(
        selection_cards, # [0] accordion
        width_policy='min',
    )

def _add_multi_choice(ps_selection, names):
    ''' Add Panel Select widgets for list of names with option 'None'. 
        ps_selection is a Column to which to add selectors.
    '''
    for name in names:
        # For value or list
        ps_selection.append(
            pn.widgets.MultiSelect(
                name=name,
                sizing_mode='stretch_width',
            )
        )

def plot_starter(callback):
    ''' Create a row with a Plot button and spinner with a button callback to start spinner. '''
    plot_button = pn.widgets.Button(
        name='Plot',
        button_style='outline',
        button_type='primary',
        sizing_mode='fixed',
        width=100,
    )

    plot_spinner = pn.indicators.LoadingSpinner(
        value=False,
        size=25,
    )

    start_spinner = pn.bind(callback, plot_button)

    return pn.Row(
        plot_button,  # [0]
        plot_spinner, # [1]
        start_spinner,
    )



'''
Functions to create a raster xarray Dataset from xradio ProcessingSet after applying plot inputs
'''

import numpy as np

from xradio.measurement_set._utils._utils.stokes_types import stokes_types

from vidavis.data.measurement_set.processing_set._ps_concat import concat_ps_xdt
from vidavis.data.measurement_set.processing_set._ps_coords import set_datetime_coordinate
from vidavis.data.measurement_set.processing_set._ps_select import select_ms
from vidavis.data.measurement_set.processing_set._xds_data import get_correlated_data, get_axis_data

def raster_data(ps_xdt, plot_inputs, logger):
    '''
    Create raster xds: y_axis vs x_axis for vis axis.
        ps_xdt (xarray DataTree): input datasets.
        plot_inputs (dict): user inputs for plot
        logger (graphviper logger): logger
    Returns: selected xarray Dataset of visibility component and updated selection
    '''
    raster_xdt = _select_raster_dimensions(ps_xdt, plot_inputs, logger)

    # Create xds from concat ms_xds in ps
    raster_xds = concat_ps_xdt(raster_xdt, logger)

    data_group = plot_inputs['data_group']
    correlated_data = get_correlated_data(raster_xds, data_group)
    if raster_xds[correlated_data].count() == 0:
        raise RuntimeError("Plot failed: raster plane selection yielded data with all nan values.")

    # Set complex component of vis data
    raster_xds[correlated_data] = get_axis_data(raster_xds,
        plot_inputs['vis_axis'],
        data_group
    )

    # Convert float time to datetime
    set_datetime_coordinate(raster_xds)

    # Apply aggregator
    raster_xds = aggregate_data(raster_xds, plot_inputs, logger)

    logger.debug(f"Plotting visibility data with shape: {raster_xds[correlated_data].shape}")
    return raster_xds

def _select_ms(ps_xdt, logger, **selection):
    ''' Select ProcessingSet MeasurementSets for raster data '''
    return select_ms(ps_xdt, logger, indexers=None, method=None, tolerance=None, drop=False, **selection)

def _select_raster_dimensions(ps_xdt, plot_inputs, logger):
    ''' Select default dimensions if needed for raster data '''
    # Determine which dims must be selected, add to selection, and do selection
    dims_to_select = _get_raster_selection_dims(plot_inputs)
    if not dims_to_select:
        return ps_xdt

    selection = plot_inputs['selection'] if 'selection' in plot_inputs else None
    dim_selection = {}

    for dim in dims_to_select:
        # Select first value (by index) and add to dim selection, or apply iter_axis value
        if not selection or dim not in selection:
            dim_selection[dim] = _get_first_dim_value(ps_xdt, dim, plot_inputs, logger)
        elif 'iter_axis' in plot_inputs and dim == plot_inputs['iter_axis']:
            dim_selection[dim] = selection[dim]
    if dim_selection:
        # Remove from selection for next plot
        if 'iter_axis' in plot_inputs and plot_inputs['iter_axis']:
            selection.pop(plot_inputs['iter_axis'])
        logger.info(f"Applying raster plane selection (using first index or iter value): {dim_selection}")
        plot_inputs['dim_selection'] = dim_selection
        return _select_ms(ps_xdt, logger, **dim_selection)
    return ps_xdt

def _get_raster_selection_dims(plot_inputs):
    ''' Return which dimensions should be selected for raster plot.
        List of dimensions which are not x, y, or agg axis. '''
    data_dims = plot_inputs['data_dims'].copy()
    if plot_inputs['x_axis'] in data_dims:
        data_dims.remove(plot_inputs['x_axis'])
    if plot_inputs['y_axis'] in data_dims:
        data_dims.remove(plot_inputs['y_axis'])
    if plot_inputs['aggregator'] and plot_inputs['agg_axis']:
        for axis in plot_inputs['agg_axis']:
            data_dims.remove(axis)
    return data_dims

def _get_first_dim_value(ps_xdt, dim, plot_inputs, logger):
    ''' Return value of first dimension by index for polarization or by value for others. '''
    # If iter_axis, get first dim value after iter value is selected to avoid empty selected ps
    iter_ps = ps_xdt
    if 'iter_axis' in plot_inputs and plot_inputs['iter_axis']:
        iter_axis = plot_inputs['iter_axis']
        iter_selection = {iter_axis: plot_inputs['selection'][iter_axis]}
        logger.debug(f"Applying {iter_axis} iter_axis selection to select first {dim} value")
        iter_ps = _select_ms(ps_xdt, logger, **iter_selection)

    values = []
    if dim == "polarization":
        # Get sorted list of polarization ids used
        stokes = list(stokes_types.values())
        for ms_xdt in iter_ps.values():
            for val in ms_xdt.polarization.values:
                values.append(stokes.index(val))
        sorted_values = sorted(list(set(values)))
        # Get pol name for pol id
        return stokes[sorted_values[0]]

    if dim == 'baseline':
        for ms_xdt in iter_ps.values():
            ant1_names = ms_xdt.baseline_antenna1_name.values
            ant2_names = ms_xdt.baseline_antenna2_name.values
            for baseline_id in ms_xdt.baseline_id.values:
                values.append(ant1_names[baseline_id] + " & " + ant2_names[baseline_id])
            return sorted(list(set(values)))[0]

    # Get sorted values list
    for ms_xdt in iter_ps.values():
        values.extend(ms_xdt[dim].values.tolist())
    return sorted(list(set(values)))[0]

def _add_index_dimensions(xds):
    ''' Add index coordinates for plotting if coord has dimension '''
    # Baseline/antenna id
    if "baseline" in xds.coords and xds.baseline.dims:
        xds = xds.assign_coords({"baseline_id": (xds.baseline.dims, np.array(range(xds.baseline.size)))})
        xds = xds.swap_dims({"baseline": "baseline_id"})
    elif "antenna_name" in xds.coords and xds.antenna_name.dims:
        xds = xds.assign_coords({"antenna_id": (xds.antenna_name.dims, np.array(range(xds.antenna_name.size)))})

    # Polarization id
    if "polarization" in xds.coords and xds.polarization.dims:
        xds = xds.assign_coords({"polarization_id": (xds.polarization.dims, np.array(range(xds.polarization.size)))})

    return xds

def aggregate_data(xds, plot_inputs, logger):
    ''' Apply aggregator to agg axis list. '''
    if not plot_inputs['aggregator']:
        return xds

    aggregator = plot_inputs['aggregator']
    agg_axis = plot_inputs['agg_axis']
    agg_xds = xds

    # Check if agg axes have been selected (selection or iteration) and are no longer a dimension
    apply_agg_axis = [axis for axis in agg_axis if axis in xds.dims]
    logger.debug(f"Applying {aggregator} to {apply_agg_axis}.")

    if aggregator == 'max':
        agg_xds = xds.max(dim=apply_agg_axis, keep_attrs=True)
    elif aggregator == 'mean':
        agg_xds = xds.mean(dim=apply_agg_axis, keep_attrs=True)
    elif aggregator == 'median':
        agg_xds = xds.median(dim=apply_agg_axis, keep_attrs=True)
    elif aggregator == 'min':
        agg_xds = xds.min(dim=apply_agg_axis, keep_attrs=True)
    elif aggregator == 'std':
        agg_xds = xds.std(dim=apply_agg_axis, keep_attrs=True)
    elif aggregator == 'sum':
        agg_xds = xds.sum(dim=apply_agg_axis, keep_attrs=True)
    elif aggregator == 'var':
        agg_xds = xds.var(dim=apply_agg_axis, keep_attrs=True)
    return agg_xds



'''
Functions for plot axes labels
'''

# pylint: disable=inconsistent-return-statements
def get_coordinate_labels(xds, coordinate):
    ''' Return coordinate values as string or list of strings, or None if numeric '''
    if coordinate == 'time':
        return _get_time_labels(xds.time)
    if coordinate == 'baseline':
        return _get_baseline_antenna_labels(xds.baseline)
    if coordinate == 'antenna_name':
        return _get_baseline_antenna_labels(xds.antenna_name)
    if coordinate == 'frequency':
        return _get_frequency_labels(xds.frequency)
    if coordinate == 'polarization':
        return _get_polarization_labels(xds.polarization)
# pylint: enable=inconsistent-return-statements

def get_axis_labels(xds, axis):
    ''' Return axis name, label and ticks, reindex for regular axis '''
    labels = get_coordinate_labels(xds, axis)
    ticks = list(enumerate(labels)) if labels is not None and isinstance(labels, list) else None

    if axis == "time":
        start_date, _ = _get_date_range(xds.time)
        label =  f"Time {xds.time.attrs['scale'].upper()} ({start_date})"
    elif axis == "baseline":
        label = "Baseline Antenna1"
        ticks = _get_baseline_ant1_ticks(ticks)
    elif axis == "antenna_name":
        label = "Antenna"
    elif axis == "frequency":
        label = f"Frequency ({xds.frequency.attrs['units']}) {xds.frequency.attrs['observer'].upper()}"
    elif axis == "polarization":
        label =  "Polarization"
    else:
        label = None
    return (axis, label, ticks)

def get_vis_axis_labels(xds, data_group, correlated_data, vis_axis, include_unit=True):
    ''' Get vis axis label for colorbar. Returns (axis, label, ticks) '''
    label = vis_axis.capitalize()
    if data_group != 'base':
        label += f":{data_group.capitalize()}"

    if include_unit:
        units = None
        if vis_axis in xds.data_vars and 'units' in xds[vis_axis].attrs:
            units = xds[vis_axis].units
        else:
            if 'units' in xds[correlated_data].attrs:
                units = xds[correlated_data].units
        if units:
            label += f" ({units})"

    return (vis_axis, label)

def _get_time_labels(time_xda):
    ''' Return time as formatted string, or None to autogenerate ticks '''
    if time_xda.size == 1:
        time = time_xda.values.astype('datetime64[s]').item().strftime("%d-%b-%Y %H:%M:%S")
        time += " " + time_xda.attrs['scale'].upper()
        return time
    return None # auto ticks from time values

def _get_baseline_antenna_labels(baseline_antenna_xda):
    ''' Return baseline pairs as string or list of strings '''
    if baseline_antenna_xda.size == 1:
        return baseline_antenna_xda.values
    return baseline_antenna_xda.values.ravel().tolist()

def _get_polarization_labels(polarization_xda):
    ''' Return polarization as single string or list of strings '''
    if polarization_xda.size == 1: # string
        return polarization_xda.values
    return list(polarization_xda.values) # array of strings

def _get_frequency_labels(frequency_xda):
    ''' Return frequency as formatted string, or None to autogenerate ticks '''
    if frequency_xda.size == 1:
        return f"{frequency_xda.item():.4f} {frequency_xda.attrs['units']} {frequency_xda.attrs['observer'].upper()}"
    return None # auto ticks from frequency values

def _get_date_range(time_xda):
    ''' Return date as dd-Mon-yyyy e.g. 23-Aug-2010 '''
    if time_xda.size == 1:
        date = time_xda.values.astype('datetime64[s]').item().strftime("%d-%b-%Y")
        return (date, date)
    start_date = time_xda.values[0].astype('datetime64[s]').item().strftime("%d-%b-%Y")
    end_date = time_xda.values[time_xda.size - 1].astype('datetime64[s]').item().strftime("%d-%b-%Y")
    return (start_date, end_date)

def _get_baseline_ant1_ticks(baseline_ticks):
    ''' Return labels for each new ant1 name in baselines '''
    ant1_ticks = []
    last_ant1 = None
    last_idx = None

    # space by minimum increment to avoid overlapping tick labels
    min_increment = max(int(len(baseline_ticks) / 50), 1)

    for idx, tick in baseline_ticks:
        ant1_name = tick.split(' & ')[0]
        if ant1_name != last_ant1:
            last_ant1 = ant1_name
            if last_idx is None or ((idx - last_idx) >= min_increment):
                ant1_ticks.append((idx, ant1_name))
                last_idx = idx
    return ant1_ticks



########################################################################
#
# Copyright (C) 2021,2023
# Associated Universities, Inc. Washington DC, USA.
#
# This script is free software; you can redistribute it and/or modify it
# under the terms of the GNU Library General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
# License for more details.
#
# You should have received a copy of the GNU Library General Public License
# along with this library; if not, write to the Free Software Foundation,
# Inc., 675 Massachusetts Ave, Cambridge, MA 02139, USA.
#
# Correspondence concerning AIPS++ should be adressed as follows:
#        Internet email: casa-feedback@nrao.edu.
#        Postal address: AIPS++ Project Office
#                        National Radio Astronomy Observatory
#                        520 Edgemont Road
#                        Charlottesville, VA 22903-2475 USA
#
########################################################################
'''Contains functions for attaching static value as attributes of functions'''

def static_vars(**kwargs):
    '''Initialize static function variables to for use within a function.

    This function is used as a decorator which allows for the initialization of
    static local variables for use within a function. It is used like:

            @static_vars(counter=0)
            def foo():
                foo.counter += 1
                print "Counter is %d" % foo.counter

    This is used from:
    https://stackoverflow.com/questions/279561/what-is-the-python-equivalent-of-static-variables-inside-a-function?rq=1

    Parameters
    ----------
    Initialized static local variables.
    '''

    def decorate(func):
        for k, v in kwargs.items():
            setattr(func, k, v)
        return func

    return decorate


def static_dir(func):
    '''return a list of static variables associated with ``func``

    Parameters
    ----------
    func: function
        function with static variables
    '''
    return [a for a in dir(func) if a[0] != '_']



''' Apply selection dict to ProcessingSet and MeasurementSetXds '''

from pandas import to_datetime

import xarray as xr

from vidavis.plot.ms_plot._ms_plot_constants import TIME_FORMAT

def select_ps(ps_xdt, logger, query=None, string_exact_match=True, **kwargs):
    '''
        Apply selection query and kwargs to ProcessingSet using exact match or partial match.
        See https://xradio.readthedocs.io/en/latest/measurement_set/schema_and_api/measurement_set_api.html#xradio.measurement_set.ProcessingSetXdt.query
        Select Processing Set first (ps summary columns), then each MeasurementSetXds, where applicable.
        Returns selected ProcessingSet DataTree.
        Throws exception if selection fails.
    '''
    # Do PSXdt selection
    logger.debug(f"Applying selection to ProcessingSet: query={query}, {kwargs}")
    ps_selected_xdt = ps_xdt.xr_ps.query(query=query, string_exact_match=string_exact_match, **kwargs)
    if string_exact_match:
        ps_selected_xdt = _select_ps_ms(ps_selected_xdt, kwargs)
    ps_selected_xdt.attrs = ps_xdt.attrs
    return ps_selected_xdt

#pylint: disable=too-many-arguments, too-many-positional-arguments, too-many-locals
def select_ms(ps_xdt, logger, indexers=None, method=None, tolerance=None, drop=False, **indexers_kwargs):
    ''' Apply selection to each MeasurementSetXdt.
        See https://xradio.readthedocs.io/en/latest/measurement_set/schema_and_api/measurement_set_api.html#xradio.measurement_set.MeasurementSetXdt.sel
        Return selected ProcessingSet DataTree.
        Throws exception if selection fails.
    '''
    # Sort out selection by dimension and baseline keywords; xr_ms can only select dims
    dim_selection = None
    time_selection = None
    baseline_selection = None
    if indexers_kwargs:
        logger.debug(f"Applying selection to each MeasurementSet: {indexers_kwargs}")
        dim_selection, time_selection, baseline_selection = _sort_selections(ps_xdt, indexers_kwargs)

    # Report (debug) selected numeric values, possibly using 'nearest' or other method
    selected_ps = xr.DataTree() # return value

    for name, ms_xdt in ps_xdt.items():
        success = True
        try:
            ms_xdt = ms_xdt.xr_ms.sel(indexers=indexers, method=method, tolerance=tolerance, drop=drop, **dim_selection)

            if time_selection:
                ms_xdt, success = _select_time(ms_xdt, time_selection, method, tolerance, drop)

            if success and baseline_selection:
                ms_xdt, success = _select_baseline(ms_xdt, baseline_selection)
        except KeyError:
            # Not in this MS
            success = False

        if success:
            selected_ps[name] = ms_xdt

    if len(selected_ps.keys()) == 0:
        raise KeyError("MeasurementSet selection yielded empty ProcessingSet.")

    selected_ps.attrs = ps_xdt.attrs
    return selected_ps
#pylint: enable=too-many-arguments, too-many-positional-arguments, too-many-locals

def _select_ps_ms(ps_xdt, ps_selection):
    ''' Select MeasurementSets in ProcessingSet according to ps selection.
        Raises exception if selection fails.
    '''
    ms_selected_xdt = xr.DataTree() # return value
    ms_selection = {}
    for key, val in ps_selection.items():
        if key in ['field_name', 'scan_name', 'polarization']:
            ms_selection[key] = val

    # Do selection on each ms_xdt
    for name, ms_xdt in ps_xdt.items():
        # Include ms only if all selections succeed
        success = True

        for key, val in ms_selection.items():
            try:
                if key == 'polarization':
                    ms_xdt = ms_xdt.sel(polarization=val)
                elif key in ['scan_name', 'field_name', 'source_name']:
                    ms_xdt, success = _select_time_dim_coordinate(ms_xdt, key, val)
                    if not success:
                        break
            except KeyError:
                # Selection failed
                success = False
                break
        if success:
            ms_selected_xdt[name] = ms_xdt

    if len(ms_selected_xdt.keys()) == 0:
        raise KeyError("Selection yielded no MeasurementSets in ProcessingSet.")

    return ms_selected_xdt

def _select_time_dim_coordinate(ms_xdt, coordinate, selection):
    ''' Select MeasurementSet coordinate with time dimension. Returns selected ms_xdt and success. '''
    success = False
    times = []
    time_xda = ms_xdt.time

    if isinstance(selection, str):
        selection = [selection]

    # Find times for selection
    for coord_sel in selection:
        sel_time_xda = time_xda.sel(time=time_xda[coordinate]==coord_sel)
        if sel_time_xda.size > 0:
            times.extend(sel_time_xda.values.tolist())

    # Select time in ms_xdt
    if len(times) == 0:
        return ms_xdt, success

    success = True
    if len(times) == 1:
        ms_xdt = ms_xdt.sel(time=times[0])
    else:
        ms_xdt = ms_xdt.sel(time=times)
    return ms_xdt, success

def _sort_selections(ps, selection):
    ''' Separate selection into dimension, time, and baseline/antenna selections '''
    dimension_keys = list(ps.xr_ps.get_max_dims().keys()) # includes 'time'
    baseline_keys = ['antenna1', 'antenna2', 'baseline']

    dim_selection = {}
    time_selection = {}
    baseline_selection = {}

    for key, val in selection.items():
        if key == 'time':
            # Convert time strings to float timestamps
            if isinstance(val, float):
                time_selection['time'] = val
            elif isinstance(val, str):
                time_selection['time'] = to_datetime(val).timestamp()
            elif isinstance(val, list):
                times = []
                for time_sel in val:
                    times.append(to_datetime(time_sel).timestamp())
                time_selection['time'] = times
            elif isinstance(val, slice):
                time_selection['time'] = slice(to_datetime(val.start).timestamp(), to_datetime(val.stop).timestamp(), val.step)
            else:
                raise TypeError(f"Time selection {selection} must be string, list, or slice")
        elif key in dimension_keys:
            dim_selection[key] = val
        elif key in baseline_keys:
            baseline_selection[key] = val
    return dim_selection, time_selection, baseline_selection

def _get_values_for_time_selection(ms_xdt, selection, method, tolerance, drop):
    ''' Return list of times in time_list which can be selected from MeasurementSet. '''
    ms_vals = []
    for val in selection['time']:
        try:
            ms_selection = {'time': val}
            ms_xdt.xr_ms.sel(indexers=None, method=method, tolerance=tolerance, drop=drop, **ms_selection)
            # Selection succeeded, include this value
            ms_vals.append(val)
        except KeyError:
            # Selection failed, do not include value
            continue
    return ms_vals

def _select_time(ms_xdt, selection, method, tolerance, drop):
    ''' Select MeasurementSet time dimension.
        Return selected ms_xdt and whether selection succeeded.
    '''
    if isinstance(selection['time'], slice):
        time_method = None
        time_tolerance = None
    else:
        # set method and tolerance for time value and list selection
        time_method = 'nearest' if method is None else method
        time_tolerance = ms_xdt.time.attrs['integration_time']['data'] if tolerance is None else tolerance

    if isinstance(selection['time'], list):
        # Only use times which exist in this ms, since sel() fails if all items in list cannot be selected.
        ms_times = _get_values_for_time_selection(ms_xdt, selection, time_method, time_tolerance, drop)
        if not ms_times:
            return ms_xdt, False
        ms_time_selection = {'time': ms_times}
        ms_xdt = ms_xdt.xr_ms.sel(indexers=None, method=time_method, tolerance=time_tolerance, drop=drop, **ms_time_selection)
    else:
        # Select str or slice
        ms_xdt = ms_xdt.xr_ms.sel(indexers=None, method=time_method, tolerance=time_tolerance, drop=drop, **selection)
        if ms_xdt.time.size == 0:
            return ms_xdt, False
    return ms_xdt, True

def _get_ms_selected_times(ms_xdt):
    ''' Return list of nearest times selected from ms as time strings '''
    time_attrs = ms_xdt.time.attrs
    try:
        return to_datetime(ms_xdt.time, unit=time_attrs['units'][0], origin=time_attrs['format']).strftime(TIME_FORMAT).values
    except AttributeError:
        return [to_datetime(ms_xdt.time, unit=time_attrs['units'][0], origin=time_attrs['format']).strftime(TIME_FORMAT)]

def _select_baseline(ms_xdt, selection):
    ''' Select MeasurementSet baseline/antenna coordinates.
        Return selected ms_xdt and whether selection succeeded.
    '''
    baseline_ids = []
    baseline_xda = ms_xdt.baseline_id

    for key, val in selection.items():
        if isinstance(val, str):
            baseline_ids.extend(_get_baseline_ids(baseline_xda, key, val))
        elif isinstance(val, list):
            for item in val:
                baseline_ids.extend(_get_baseline_ids(baseline_xda, key, item))
        else:
            raise TypeError("Can only select baselines and antennas by str or list")

    success = len(baseline_ids) > 0
    if success:
        baseline_ids = sorted(list(set(baseline_ids)))
        if len(baseline_ids) == 1:
            ms_xdt = ms_xdt.sel(baseline_id=baseline_ids[0])
        else:
            ms_xdt = ms_xdt.sel(baseline_id=baseline_ids)
    return ms_xdt, success

def _get_baseline_ids(baseline_xda, key, val):
    ''' Return list of baseline_ids for baseline or antenna selections. '''
    if not isinstance(val, str):
        raise TypeError("baseline/antenna selection value must be str")

    antenna1 = None
    antenna2 = None
    if key == 'baseline':
        ant1, ant2 = val.split('&')
        antenna1 = ant1.strip()
        antenna2 = ant2.strip()
    elif key == 'antenna1':
        antenna1 = val
    elif key == 'antenna2':
        antenna2 = val

    try:
        sel_baseline_xda = None
        # Select antenna1
        if antenna1 is not None:
            sel_baseline_xda = baseline_xda.sel(baseline_id=baseline_xda.baseline_antenna1_name==antenna1)

        # Select antenna2
        if antenna2 is not None:
            if sel_baseline_xda is None:
                sel_baseline_xda = baseline_xda.sel(baseline_id=baseline_xda.baseline_antenna2_name==antenna2)
            else:
                sel_baseline_xda = sel_baseline_xda.sel(baseline_id=sel_baseline_xda.baseline_antenna2_name==antenna2)

        # Return baseline ids
        if sel_baseline_xda is not None and sel_baseline_xda.size > 0:
            return sel_baseline_xda.values.tolist()
        return []
    except KeyError:
        return []



'''
   Calculate statistics on xradio ProcessingSet data.
'''

import dask
import numpy as np

from xradio.measurement_set.load_processing_set import ProcessingSetIterator
from graphviper.graph_tools import generate_dask_workflow
from graphviper.graph_tools.coordinate_utils import make_parallel_coord, interpolate_data_coords_onto_parallel_coords
from graphviper.graph_tools.map import map as graph_map
from graphviper.graph_tools.reduce import reduce as graph_reduce

try:
    from toolviper.dask.client import get_client
    _HAVE_TOOLVIPER = True
except ImportError:
    _HAVE_TOOLVIPER = False

from vidavis.data.measurement_set.processing_set._xds_data import get_correlated_data, get_axis_data

def calculate_ps_stats(ps_xdt, ps_store, vis_axis, data_group, logger):
    '''
        Calculate stats for unflagged visibilities: min, max, mean, std
        ps_xdt (xarray.DataTree): input MeasurementSet opened from zarr file
        ps_store (str): path to visibility zarr file
        vis_axis (str): complex component (amp, phase, real, imag)
        Returns: stats tuple (min, max, mean, stddev) or None if all data flagged (count=0)
    '''
    input_params = {}
    input_params['input_data_store'] = ps_store
    input_params['xdt'] = ps_xdt
    input_params['vis_axis'] = vis_axis
    input_params['data_group'] = data_group
    for ms_xdt in ps_xdt.values():
        if data_group in ms_xdt.attrs['data_groups']:
            input_params['correlated_data'] = get_correlated_data(ms_xdt.ds, data_group)
            break

    if _HAVE_TOOLVIPER:
        active_client = get_client() # could be None if not started before using application
    else:
        active_client = None
    n_threads = active_client.thread_info()['n_threads'] if active_client is not None else 4
    logger.debug(f"Setting {n_threads} n_chunks for parallel coords.")
    mapping = _get_task_data_mapping(ps_xdt, n_threads)

    data_min, data_max, data_mean = _calc_basic_stats(ps_xdt, mapping, input_params, logger)
    if np.isfinite(data_mean):
        input_params['mean'] = data_mean
        data_stddev = _calc_stddev(ps_xdt, mapping, input_params, logger)
        return data_min, data_max, data_mean, data_stddev
    return None

def _get_task_data_mapping(ps_xdt, n_threads):
    frequencies = ps_xdt.xr_ps.get_freq_axis()
    parallel_coords = {"frequency": make_parallel_coord(coord=frequencies, n_chunks=n_threads)}
    return interpolate_data_coords_onto_parallel_coords(parallel_coords, ps_xdt)

def _calc_basic_stats(ps_xdt, mapping, input_params, logger):
    ''' Calculate min, max, mean using graph map/reduce '''
    graph = graph_map(
        input_data=ps_xdt,
        node_task_data_mapping=mapping,
        node_task=_map_stats,
        input_params=input_params
    )
    reduce_map = graph_reduce(
        graph, _reduce_stats, input_params, mode='tree'
    )
    dask_graph = generate_dask_workflow(reduce_map)
    #dask_graph.visualize(filename='stats.png')
    results = dask.compute(dask_graph)

    data_min, data_max, data_sum, data_count = results[0]
    if data_count == 0.0:
        logger.debug("stats: no unflagged data")
        return (data_min, data_max, np.inf)

    data_mean = data_sum / data_count
    logger.debug(f"basic stats: min={data_min:.4f}, max={data_max:.4f}, sum={data_sum:.4f}, count={data_count} mean={data_mean:.4f}")
    return data_min, data_max, data_mean

def _calc_stddev(ps_xdt, mapping, input_params, logger):
    ''' Calculate stddev using graph map/reduce '''
    graph = graph_map(
        input_data=ps_xdt,
        node_task_data_mapping=mapping,
        node_task=_map_variance,
        input_params=input_params
    )
    reduce_map = graph_reduce(
        graph, _reduce_variance, input_params, mode='tree'
    )
    dask_graph = generate_dask_workflow(reduce_map)
    results = dask.compute(dask_graph)

    var_sum, var_count = results[0]
    data_variance = var_sum / var_count
    data_stddev = data_variance ** 0.5
    logger.debug(f"stats: variance={data_variance:.4f}, stddev={data_stddev:.4f}")
    return data_stddev

def _get_stats_xda(xds, vis_axis, data_group):
    ''' Return xda with only unflagged cross-corr visibility data '''
    # apply flags to get unflagged vis data
    xda = get_axis_data(xds, vis_axis, data_group)
    unflagged_xda = xda.where(np.logical_not(xds.FLAG))

    if unflagged_xda.count() > 0 and "baseline_antenna1_name" in unflagged_xda.coords:
        # if unflagged data, remove autocorrelation baselines
        stats_xda = unflagged_xda.where(
            unflagged_xda.baseline_antenna1_name != unflagged_xda.baseline_antenna2_name
        )
        if stats_xda.count() > 0:
            # return xda with nan where flagged or auto-corr
            return stats_xda

    # return xda with nan where flagged
    return unflagged_xda

def _map_stats(input_params):
    ''' Return min, max, sum, and count of data chunk '''
    vis_axis = input_params['vis_axis']
    data_group = input_params['data_group']
    correlated_data = input_params['correlated_data']
    min_vals = []
    max_vals = []
    sum_vals = []
    count_vals = []

    ps_iter = ProcessingSetIterator(
        input_params['data_selection'],
        input_params['input_data_store'],
        input_params['xdt'],
        input_params['data_group'],
        include_variables=[correlated_data, 'FLAG'],
        load_sub_datasets=False
    )

    for xds in ps_iter:
        xda = _get_stats_xda(xds, vis_axis, data_group)
        if xda.count() > 0:
            xda_data = xda.values.ravel()
            try:
                min_vals.append(xda_data[np.nanargmin(xda_data)])
            except ValueError:
                pass
            try:
                max_vals.append(xda_data[np.nanargmax(xda_data)])
            except ValueError:
                pass
            sum_vals.append(np.nansum(xda))
            count_vals.append(xda.count().values)
    try:
        min_value = np.nanmin(min_vals)
    except ValueError:
        min_value = np.nan
    try:
        max_value = np.nanmax(max_vals)
    except ValueError:
        max_value = np.nan
    return (min_value, max_value, sum(sum_vals), sum(count_vals))

# pylint: disable=unused-argument
def _reduce_stats(graph_inputs, input_params):
    ''' Compute min, max, sum, and count of all data.
        input_parameters seems to be required although unused. '''
    data_min = 0.0
    mins = [values[0] for values in graph_inputs]
    if not np.isnan(mins).all():
        data_min = min(0.0, np.nanmin(mins))

    data_max = 0.0
    maxs = [values[1] for values in graph_inputs]
    if not np.isnan(maxs).all():
        data_max = max(0.0, np.nanmax(maxs))

    data_sum = sum(values[2] for values in graph_inputs)
    data_count = sum(values[3] for values in graph_inputs)
    return (data_min, data_max, data_sum, data_count)
# pylint: enable=unused-argument

def _map_variance(input_params):
    ''' Return sum, count, of (xda - mean) squared '''
    vis_axis = input_params['vis_axis']
    data_group = input_params['data_group']
    correlated_data = input_params['correlated_data']
    mean = input_params['mean']

    sq_diff_sum = 0.0
    sq_diff_count = 0

    ps_iter = ProcessingSetIterator(
        input_params['data_selection'],
        input_params['input_data_store'],
        input_params['xdt'],
        input_params['data_group'],
        include_variables=[correlated_data, 'FLAG'],
        load_sub_datasets=False
    )

    for xds in ps_iter:
        xda = _get_stats_xda(xds, vis_axis, data_group)
        if xda.size > 0:
            sq_diff = (xda - mean) ** 2
            sq_diff_sum += np.nansum(sq_diff)
            sq_diff_count += sq_diff.count().values
    return (sq_diff_sum, sq_diff_count)

# pylint: disable=unused-argument
def _reduce_variance(graph_inputs, input_params):
    ''' Compute sum and count of all (xda-mean) squared data.
        input_parameters seems to be required although unused. '''
    sq_diff_sum = sum(values[0] for values in graph_inputs)
    sq_diff_count = sum(values[1] for values in graph_inputs)
    return (sq_diff_sum, sq_diff_count)
# pylint: enable=unused-argument



'''
Implementation of the ``MsRaster`` application for measurement set raster plotting and editing
'''

import time

from bokeh.models.formatters import NumeralTickFormatter
import holoviews as hv
import numpy as np
import panel as pn
from pandas import to_datetime

from vidavis.bokeh._palette import available_palettes
from vidavis.plot.ms_plot._time_ticks import get_time_formatter
from vidavis.plot.ms_plot._ms_plot import MsPlot
from vidavis.plot.ms_plot._ms_plot_constants import VIS_AXIS_OPTIONS, SPECTRUM_AXIS_OPTIONS, PS_SELECTION_OPTIONS, MS_SELECTION_OPTIONS
from vidavis.plot.ms_plot._ms_plot_selectors import (file_selector, title_selector, style_selector, axis_selector,
aggregation_selector, iteration_selector, selection_selector, plot_starter)
from vidavis.plot.ms_plot._raster_plot_inputs import check_inputs
from vidavis.plot.ms_plot._raster_plot import RasterPlot

class MsRaster(MsPlot):
    '''
    Plot MeasurementSet data as raster plot.

    Args:
        ms (str, None): path to MSv2 (.ms) or MSv4 (.zarr) file. Default None. Required when show_gui=False.
        log_level (str): logging threshold. Options include 'debug', 'info', 'warning', 'error', 'critical'. Default 'info'.
        log_to_file (bool): whether to write log messages to log file "msraster-<timestamp>.log". Default True.
        show_gui (bool): whether to launch the interactive GUI in a browser tab. Default False.

    Example:
        from vidavis.apps import MsRaster
        msr = MsRaster(ms='myvis.ms')
        msr.summary()
        msr.set_style_params(unflagged_cmap='Plasma', flagged_cmap='Greys', show_colorbar=True)
        msr.select_ps(intents='OBSERVE_TARGET#ON_SOURCE')
        msr.plot(x_axis='frequency', y_axis='time', vis_axis='amp', data_group='base')
        msr.show()
        msr.save() # saves as {ms name}_raster.png
    '''

    def __init__(self, ms=None, log_level="info", log_to_file=True, show_gui=False):
        super().__init__(ms, log_level, log_to_file, show_gui, "MsRaster")
        self._raster_plot = RasterPlot()

        # Calculations for color limits
        self._spw_stats = {}
        self._spw_color_limits = {}

        if show_gui:
            # GUI based on panel widgets
            self._gui_layout = None

            # Check if plot inputs changed and new plot is needed.
            # GUI can change subparams which do not change plot
            self._last_plot_inputs = None
            self._last_style_inputs = None
            # Last plot when no new plot created (plot inputs same) or is iter Layout plot (opened in tab)
            self._last_gui_plot = None

            # Return plot for gui DynamicMap:
            # Empty plot when ms not set or plot fails
            self._empty_plot = self._create_empty_plot()

            # Set default style and plot inputs to use when launching gui
            self.set_style_params()
            self.plot()
            self._launch_gui()

            # Set filename TextInput to input ms to trigger plot
            if 'ms' in self._ms_info and self._ms_info['ms']:
                self._set_filename([self._ms_info['ms']]) # function expects list

    def colormaps(self):
        ''' List available colormap (Bokeh palettes). '''
        return available_palettes()

    def set_style_params(self, unflagged_cmap='Viridis', flagged_cmap='Reds', show_colorbar=True, show_flagged_colorbar=True):
        '''
            Set styling parameters for the plot, such as colormaps and whether to show colorbar.
            Placeholder for future styling such as fonts.

            Args:
                unflagged_cmap (str): colormap to use for unflagged data.
                flagged_cmap (str): colormap to use for flagged data.
                show_colorbar (bool): Whether to show colorbar with plot.  Default True.
        '''
        cmaps = self.colormaps()
        if unflagged_cmap not in cmaps:
            raise ValueError(f"{unflagged_cmap} not in colormaps list: {cmaps}")
        if flagged_cmap not in cmaps:
            raise ValueError(f"{flagged_cmap} not in colormaps list: {cmaps}")
        self._raster_plot.set_style_params(unflagged_cmap, flagged_cmap, show_colorbar, show_flagged_colorbar)

    def select_ps(self, string_exact_match=True, query=None, **kwargs):
        '''
        Select a subset of ProcessingSet MeasurementSets using a Pandas query or summary column names and values.
            string_exact_match (bool): whether to require exact matches for string and string list columns (default True) or partial matches (False).
            query (str): a Pandas query string to apply additional filtering.
            **kwargs (dict): keyword arguments representing summary column names and values.
        See summary(data_group) and data_groups() for selection options. Use keyword 'data_group_name' for data group selection.
        Summary column names include 'name', 'intents', 'shape', 'polarization', 'scan_name', 'spw_name',
          'field_name', 'source_name', 'field_coords', 'start_frequency', 'end_frequency'.
        The selection is applied to the data within the MeasurementSets where applicable (polarization, scan_name, field_name)
          string_exact_match=True, else the entire MS is selected.
        Selections are cumulative until clear_selection() is called.
        Raises exception with message if selection fails.

        For explanation and examples, see:
        https://xradio.readthedocs.io/en/latest/measurement_set/schema_and_api/measurement_set_api.html#xradio.measurement_set.ProcessingSetXdt.query
        '''
        if self._data and self._data.is_valid():
            try:
                self._plot_inputs['selection'] |= kwargs
                if 'data_group_name' in kwargs:
                    self._plot_inputs['data_group'] = kwargs['data_group_name']
                self._data.select_ps(query=query, string_exact_match=string_exact_match, **kwargs)
            except KeyError as ke:
                error = "ProcessingSet selection yielded empty ProcessingSet."
                if not self._show_gui:
                    error += " Modify selection or run clear_selection() to select original ProcessingSet."
                raise KeyError(error) from ke
        else:
            raise RuntimeError("Cannot select ProcessingSet: input MS path is invalid or missing.")

    def select_ms(self, indexers=None, method=None, tolerance=None, drop=False, **indexers_kwargs):
        '''
        Select MeasurementSet dimensions or data_group.
          Values may be a single value, a list, or a slice.
          Data group may be selected with "data_group_name"; see data_groups() for options.
          See get_dimension_values() for selection keywords and values.
            Dimensions may be selected with 'time', 'baseline' (visibilities), 'antenna_name' (spectrum),
              'antenna1', 'antenna2', 'frequency', and 'polarization'.
            Time selection must be in string format 'dd-Mon-YYYY HH:MM:SS' as shown in get_dimension_values('time').
        Selections are cumulative until clear_selection() is called.
        Raises exception with message if selection fails.

        For explanation of parameters and examples, see:
        https://xradio.readthedocs.io/en/latest/measurement_set/schema_and_api/measurement_set_api.html#xradio.measurement_set.MeasurementSetXdt.sel
        '''
        if self._data and self._data.is_valid():
            try:
                self._plot_inputs['selection'] |= indexers_kwargs
                self._data.select_ms(indexers=indexers, method=method, tolerance=tolerance, drop=drop, **indexers_kwargs)
                if 'data_group_name' in indexers_kwargs:
                    self._plot_inputs['data_group'] |= indexers_kwargs['data_group_name']
            except KeyError as ke:
                error = str(ke).strip("\'")
                if not self._show_gui:
                    error += " Modify selection or run clear_selection() to select original ProcessingSet."
                raise KeyError(error) from ke
        else:
            raise RuntimeError("Cannot select MeasurementSet: input MS path is invalid or missing.")

# pylint: disable=too-many-arguments, too-many-positional-arguments, too-many-locals, unused-argument
    def plot(self, x_axis='baseline', y_axis='time', vis_axis='amp', aggregator=None, agg_axis=None,
             iter_axis=None, iter_range=None, subplots=None, color_mode=None, color_range=None, title=None, clear_plots=True):
        '''
        Create a raster plot of vis_axis data.
        Plot axes include data dimensions (time, baseline/antenna, frequency, polarization).
        The first spectral window (by id) and dimensions not set as plot axes (first value) will be automatically selected if no user selection has been made, unless aggregated.

        Args:
            x_axis (str): Plot x-axis. Default 'baseline' ('antenna_name' for spectrum data).
            y_axis (str): Plot y-axis. Default 'time'.
            vis_axis (str): Complex visibility component to plot (amp, phase, real, imag). Default 'amp'.
                Call data_groups() to see options.
                MeasurementSet selection:
                    'data_group': name for correlated data, flags, weights, and uvw. Default value 'base'.
                        Use data_groups() to get data group names.
                    Dimensions:
                        Visibility dimensions: 'baseline' 'time', 'frequency', 'polarization'
                        Spectrum dimensions: 'antenna_name', 'time', 'frequency', 'polarization'
                        Default value is index 0 (after user selection) for non-axis dimensions unless aggregated.
                        Use antennas() to get antenna names. Select 'baseline' as "<name1> & <name2>".
                        Use summary() to list frequencies and polarizations.
                        TODO: how to select time?
            aggregator (str): reduction for rasterization. Default None.
                Options include 'max', 'mean', 'min', 'std', 'sum', 'var'.
            agg_axis (str, list): which dimension to apply aggregator across. Default None.
                Options include one or more dimensions.
                If agg_axis is None and aggregator is set, aggregates over all non-axis dimensions.
                If one agg_axis is selected, the non-agg dimension will be selected.
            iter_axis (str): dimension over which to iterate values (using iter_range).
            iter_range (tuple): (start, end) inclusive index values for iteration plots.
                Default (0, 0) (first iteration only). Use (0, -1) for all iterations.
                If subplots is a grid, the range is limited by the grid size.
                If subplots is a single plot, all iteration plots in the range can be saved using export_range in save().
            subplots (None, tuple): set a grid of (rows, columns).  None = (1, 1) for single plot.
                Use with iter_axis and iter_range, or clear_plots=False.
                If used in multiple calls, the last subplots tuple will be used to determine grid to show or save.
            color_mode (None, str): Whether to limit range of colorbar.  Default None (no limit).
                Options include None (use data limits), 'auto' (calculate limits for amplitude), and 'manual' (use range in color_range).
                'auto' is equivalent to None if vis_axis is not 'amp'.
                When subplots is set, the 'auto' or 'manual' range will be used for all plots.
            color_range (tuple): (min, max) of colorbar to use if color_mode is 'manual'.
            title (str): Plot title, default None (no title)
                Set title='ms' to generate title from ms name and iter_axis value, if any.
            clear_plots (bool): whether to clear list of plots. Default True.

        If not show_gui and plotting is successful, use show() or save() to view/save the plot only.
        '''
        inputs = locals() # collect arguments into dict (not unused as pylint complains!)
        if self._plot_inputs['selection']:
            self._logger.info("Create raster plot with selection: %s", self._plot_inputs['selection'])

        start = time.time()

        # Clear for new plot
        self._reset_plot(clear_plots)

        # Get data dimensions if valid MS is set to check input axes
        if 'data_dims' in self._ms_info:
            inputs['data_dims'] = self._ms_info['data_dims'] if 'data_dims' in self._ms_info else None

        # Validate input arguments then set
        check_inputs(inputs)
        self._set_plot_inputs(inputs)

        if not self._show_gui:
            # Cannot plot if no MS
            if not self._data or not self._data.is_valid():
                raise RuntimeError("Cannot plot MS: input MS path is invalid or missing.")

            # Create raster plot and add to plot list
            try:
                if self._plot_inputs['iter_axis']:
                    self._do_iter_plot(self._plot_inputs)
                else:
                    plot = self._do_plot(self._plot_inputs)
                    self._plots.append(plot)
            except RuntimeError as e:
                error = f"Plot failed: {str(e)}"
                super()._notify(error, "error", 0)

            self._logger.debug("Plot elapsed time: %.2fs.", time.time() - start)
# pylint: enable=too-many-arguments, too-many-positional-arguments, too-many-locals, unused-argument

    def save(self, filename='', fmt='auto', width=900, height=600):
        '''
        Save plot to file.

        Args:
            filename (str): Name of file to save. Default '': the plot will be saved as {ms}_raster.{ext}.
                If fmt is not set for extension, plot will be saved as .png.
            fmt (str): Format of file to save ('png', 'html', or 'gif').
                Default 'auto': inferred from filename extension.
            width (int): width of exported plot.
            height (int): height of exported plot.

        If iteration plots were created:
            If subplots is a grid, the layout plot will be saved to a single file.
            If subplots is a single plot, iteration plots will be saved individually,
                with a plot index appended to the filename: {filename}_{index}.{ext}.
        '''
        if not filename:
            filename = f"{self._ms_info['basename']}_raster.png"
        super().save(filename, fmt, width, height)

    def _set_plot_inputs(self, plot_inputs):
        ''' Set inputs from latest plot() call '''
        for key, val in plot_inputs.items():
            self._plot_inputs[key] = val

    def _do_plot(self, plot_inputs):
        ''' Create plot using plot inputs '''
        if not self._plot_init:
            self._init_plot(plot_inputs)

        # Select vis_axis data to plot and update selection; returns xarray Dataset
        raster_data = self._data.get_raster_data(plot_inputs)

        # Add params needed for plot: auto color range and ms name
        self._set_auto_color_range(plot_inputs) # set calculated limits if auto mode
        ms_name = self._ms_info['basename'] # for title
        self._raster_plot.set_plot_params(raster_data, plot_inputs, ms_name)

        # Show plot inputs in log
        super()._set_plot_params(plot_inputs | self._raster_plot.get_plot_params()['style'])
        self._logger.info("MsRaster plot parameters: %s", ", ".join(self._plot_params))

        # Make plot. Add data min/max if GUI is shown to update color limits range.
        return self._raster_plot.raster_plot(raster_data, self._logger, self._show_gui)

    def _do_iter_plot(self, plot_inputs):
        ''' Create one plot per iteration value in iter_range which fits into subplots '''
        # Default (0, 0) (first iteration only). Use (0, -1) for all iterations.
        # If subplots is a grid, end iteration index is limited by the grid size.
        # If subplots is a single plot, all iteration plots in the range can be saved using export_range in save().
        iter_axis = plot_inputs['iter_axis']
        iter_range = plot_inputs['iter_range']
        subplots = plot_inputs['subplots']

        iter_range = (0, 0) if iter_range is None else iter_range
        start_idx, end_idx = iter_range

        # Init plot before getting iter values
        self._init_plot(plot_inputs)

        iter_values = self._data.get_dimension_values(iter_axis)
        n_iter = len(iter_values)

        if start_idx >= n_iter:
            raise IndexError(f"iter_range start {start_idx} is greater than number of iterations {n_iter}")
        end_idx = n_iter if (end_idx == -1 or end_idx >= n_iter) else end_idx + 1
        num_iter_plots = end_idx - start_idx

        # Plot the minimum of iter range or subplots number of plots
        num_subplots = np.prod(subplots) if subplots else 1
        num_iter_plots = min(num_iter_plots, num_subplots) if num_subplots > 1 else num_iter_plots
        end_idx = start_idx + num_iter_plots

        # Set each iter value in ms_selection
        if 'ms_selection' not in plot_inputs:
            plot_inputs['ms_selection'] = {}

        for i in range(start_idx, end_idx):
            # Select iteration value and make plot
            value = iter_values[i]
            self._logger.info("Plot %s iteration index %s value %s", iter_axis, i, value)
            plot_inputs['selection'][iter_axis] = value
            try:
                plot = self._do_plot(plot_inputs)
                self._plots.append(plot)
            except RuntimeError as e:
                self._logger.info("Iteration plot for value %s failed: %s", str(value), str(e))
                continue

    def _init_plot(self, plot_inputs):
        ''' Apply automatic selection '''
        # Remove previous auto selections
        for key in ['dim_selection', 'auto_spw']:
            try:
                del self._plot_inputs[key]
            except KeyError:
                pass

        # Automatically select data group and spw name if not user-selected
        auto_selection = {}
        if 'data_group' not in plot_inputs:
            auto_selection['data_group_name'] = 'base'
            plot_inputs['data_group'] = 'base'
        if 'selection' not in plot_inputs or 'spw_name' not in plot_inputs['selection']:
            first_spw = self._data.get_first_spw(plot_inputs['data_group'])
            auto_selection['spw_name'] = first_spw
            plot_inputs['auto_spw'] = first_spw # do not add to user selection

        if auto_selection:
            # Do selection and save to plot inputs
            self._logger.info("Automatic selection of data group and/or spw: %s", auto_selection)
            self._data.select_ps(query=None, string_exact_match=True, **auto_selection)

        # Print data info for spw selection
        self._logger.info("Plotting %s msv4 datasets.", self._data.get_num_ms())
        self._logger.info("Maximum dimensions for selected spw: %s", self._data.get_max_data_dims())
        self._plot_init = True

    def _set_auto_color_range(self, plot_inputs):
        ''' Calculate stats for color limits for non-gui amplitude plots. '''
        color_mode = plot_inputs['color_mode']
        color_limits = None

        if color_mode == 'auto':
            if plot_inputs['vis_axis']=='amp' and not plot_inputs['aggregator']:
                # For amplitude, limit colorbar range using stored per-spw ms stats
                spw_name = self._plot_inputs['selection']['spw_name'] if ('selection' in plot_inputs and 'spw_name' in plot_inputs['selection']) else self._plot_inputs['auto_spw']
                if spw_name in self._spw_color_limits:
                    color_limits = self._spw_color_limits[spw_name]
                else:
                    # Select spw name and data group only, no dimensions
                    data_group = self._plot_inputs['data_group']
                    spw_data_selection = {'spw_name': spw_name, 'data_group_name': data_group}
                    color_limits = self._calc_amp_color_limits(spw_data_selection)
                    self._spw_color_limits[spw_name] = color_limits
        plot_inputs['auto_color_range'] = color_limits

        if color_limits:
            self._logger.info("Setting amplitude color range: (%.4f, %.4f).", color_limits[0], color_limits[1])
        elif color_mode is None:
            self._logger.info("Autoscale color range")
        else:
            self._logger.info("Using manual color range: %s", plot_inputs['color_range'])

    def _calc_amp_color_limits(self, selection):
        # Calculate colorbar limits from amplitude stats for unflagged data in selected spw
        self._logger.info("Calculating stats for colorbar limits.")
        start = time.time()

        ms_stats = self._data.get_vis_stats(selection, 'amp')
        self._spw_stats['spw_name'] = ms_stats
        if not ms_stats:
            return None # autoscale

        min_val, max_val, mean, std = ms_stats

        data_min = min(0.0, min_val)
        clip_min = max(data_min, mean - (3.0 * std))
        data_max = max(0.0, max_val)
        clip_max = min(data_max, mean + (3.0 * std))

        if clip_min == 0.0 and clip_max == 0.0:
            color_limits = None # flagged data only
        else:
            color_limits = (clip_min, clip_max)
        self._logger.debug("Stats elapsed time: %.2fs.", time.time() - start)
        return color_limits

    def _reset_plot(self, clear_plots=True):
        ''' Reset any plot settings for a new plot '''
        # Clear plot list
        if clear_plots:
            super().clear_plots()

        # Clear params set for last plot
        self._raster_plot.reset_plot_params()

        # Unitialize plot to redo auto selections if needed
        self._plot_init = False

    ### -----------------------------------------------------------------------
    ### Interactive GUI
    ### -----------------------------------------------------------------------
    def _launch_gui(self):
        ''' Use Holoviz Panel to create a dashboard for plot inputs. '''
        # Select MS
        file_selectors = file_selector('Path to MeasurementSet (ms or zarr) for plot', '~' , self._set_filename)

        # Select style - colormaps, colorbar, color limits
        style_selectors = style_selector(self._set_style_params, self._set_color_range)

        # Set title
        title_input = title_selector(self._set_title)

        # Select x, y, and vis axis
        x_axis = self._plot_inputs['x_axis']
        y_axis = self._plot_inputs['y_axis']
        data_dims = self._ms_info['data_dims'] if 'data_dims' in self._ms_info else None
        axis_selectors = axis_selector(x_axis, y_axis, data_dims, True, self._set_axes)

        # Select from ProcessingSet and MeasurementSet
        selection_selectors = selection_selector(self._set_ps_selection, self._set_ms_selection)

        # Generic axis options, updated when ms is set
        axis_options = data_dims if data_dims else []

        # Select aggregator and axes to aggregate
        agg_selectors = aggregation_selector(axis_options, self._set_aggregation)

        # Select iter_axis and iter value or range
        iter_selectors = iteration_selector(axis_options, self._set_iter_values, self._set_iteration)

        # Put user input widgets in accordion with only one card active at a time
        selectors = pn.Accordion(
            ("Select file", file_selectors),         # [0]
            ("Plot style", style_selectors),         # [1]
            ("Data Selection", selection_selectors), # [2]
            ("Plot axes", axis_selectors),           # [3]
            ("Aggregation", agg_selectors),          # [4]
            ("Iteration", iter_selectors),           # [5]
            ("Plot title", title_input),             # [6]
        )
        selectors.toggle = True

        # Plot button and spinner while plotting
        init_plot = plot_starter(self._update_plot_spinner)

        # Connect plot to filename and selector widgets
        dmap = hv.DynamicMap(
            pn.bind(
                self._update_plot,
                ms=file_selectors[0][0],
                do_plot=init_plot[0],
            ),
        )

        # Layout plot and input widgets in a row
        self._gui_layout = pn.Row(
            pn.Tabs(             # [0]
                ('Plot', dmap),               # [0]
                ('Plot Inputs', pn.Column()), # [1]
                sizing_mode='stretch_width',
            ),
            pn.Spacer(width=10), # [1]
            pn.Column(           # [2]
                pn.Spacer(height=25), # [0]
                selectors,            # [1]
                init_plot,            # [2]
                width_policy='min',
                width=400,
                sizing_mode='stretch_height',
            ),
            sizing_mode='stretch_height',
        )

        # Show gui
        # print("gui layout:", self._gui_layout) # for debugging
        self._gui_layout.show(title=self._app_name, threaded=True)

    ###
    ### Main callback to create plot if inputs changed
    ###
    def _update_plot(self, ms, do_plot):
        ''' Create plot with inputs from GUI.  Must return plot, even if empty plot, for DynamicMap. '''
        if self._toast:
            self._toast.destroy()

        self._get_selector("selectors").active = []
        if not ms:
            # Launched GUI with no MS
            return self._empty_plot

        # If not first plot, user has to click Plot button (do_plot=True).
        first_plot = not self._last_gui_plot
        if not do_plot and not first_plot:
            # Not ready to update plot yet, return last plot.
            return self._last_gui_plot

        if (self._set_ms(ms) or first_plot) and self._data and self._data.is_valid():
            # New MS set and is valid
            self._update_gui_axis_options()

        # Add ms path to detect change and make new plot
        self._plot_inputs['ms'] = ms

        # Do new plot or resend last plot
        self._reset_plot()
        self.clear_selection()
        gui_plot = None

        style_inputs = self._raster_plot.get_plot_params()['style']
        if self._inputs_changed(style_inputs):
            # First plot or changed plot
            try:
                # Check inputs from GUI then plot
                self._plot_inputs['data_dims'] = self._ms_info['data_dims']
                check_inputs(self._plot_inputs)
                if 'ps_selection' in self._plot_inputs or 'ms_selection' in self._plot_inputs:
                    self._do_gui_selection()
                gui_plot = self._do_gui_plot()
            except (ValueError, TypeError, KeyError, RuntimeError) as e:
                # Clear plot, inputs invalid
                self._notify(str(e), 'error', 0)
                gui_plot = self._empty_plot
        else:
            # Subparam values changed but not applied to plot
            gui_plot = self._last_gui_plot

        # Update plot inputs for gui
        self._set_plot_params(self._plot_inputs | style_inputs)

        # Save inputs to see if changed
        self._last_plot_inputs = self._plot_inputs.copy()
        self._last_style_inputs = style_inputs.copy()
        self._last_plot_inputs['ps_selection'] = self._plot_inputs['ps_selection'].copy()
        self._last_plot_inputs['ms_selection'] = self._plot_inputs['ms_selection'].copy()
        if first_plot and not do_plot:
            self._plot_inputs['ps_selection'].clear()
            self._plot_inputs['ms_selection'].clear()

        # Save plot if no new plot
        self._last_gui_plot = gui_plot

        # Add plot inputs, change plot button to outline, and stop spinner
        self._update_plot_inputs()
        self._update_plot_status(False)
        self._update_plot_spinner(False)

        return gui_plot

    def _inputs_changed(self, style_inputs):
        ''' Check if inputs changed and need new plot '''
        if not self._last_plot_inputs:
            return True

        for key, val in self._plot_inputs.items():
            if not self._values_equal(val, self._last_plot_inputs[key]):
                return True

        for key, val in style_inputs.items():
            if not self._values_equal(val, self._last_style_inputs[key]):
                return True
        return False

    def _values_equal(self, val1, val2):
        ''' Test if values are set and equal, or not set (cannot compare value with None) '''
        if val1 is not None and val2 is not None: # both set
            return val1 == val2
        if val1 is None and val2 is None: # both None
            return True
        return False # one set and other is None

    def _do_gui_selection(self):
        ''' Apply selections selected in GUI '''
        if self._plot_inputs['ps_selection']:
            self.select_ps(**self._plot_inputs['ps_selection'])
        if self._plot_inputs['ms_selection']:
            self.select_ms(**self._plot_inputs['ms_selection'])

    ###
    ### Create plot for DynamicMap
    ###
    def _do_gui_plot(self):
        ''' Create plot based on gui plot inputs '''
        if self._data and self._data.is_valid():
            try:
                if self._plot_inputs['iter_axis']:
                    # Make iter plot (possibly with subplots layout)
                    self._do_iter_plot(self._plot_inputs)
                    subplots = self._plot_inputs['subplots']
                    layout_plot, is_layout = super()._layout_plots(subplots)

                    if is_layout:
                        # Cannot show Layout in DynamicMap, show in new tab
                        super().show()
                        self._logger.info("Plot update complete")
                        return self._last_gui_plot
                    # Overlay raster plot for DynamicMap
                    self._logger.info("Plot update complete")
                    return layout_plot

                # Make single Overlay raster plot for DynamicMap
                plot = self._do_plot(self._plot_inputs)
                plot_params = self._raster_plot.get_plot_params()

                # Update color limits in gui with data range
                self._update_color_range(plot_params)

                # Update colorbar labels and limits
                plot.QuadMesh.I = self._set_plot_colorbar(plot.QuadMesh.I, plot_params, "flagged")
                plot.QuadMesh.II = self._set_plot_colorbar(plot.QuadMesh.II, plot_params, "unflagged")

                self._logger.info("Plot update complete")
                return plot.opts(
                    hv.opts.QuadMesh(tools=['hover'])
                )
            except RuntimeError as e:
                error = f"Plot failed: {str(e)}"
                super()._notify(error, "error", 0)

        # Make single Overlay raster plot for DynamicMap
        return self._empty_plot

    def _create_empty_plot(self):
        ''' Create empty Overlay plot for DynamicMap with colormap params and hover enabled '''
        plot_params = self._raster_plot.get_plot_params()
        plot = hv.Overlay(
            hv.QuadMesh([]).opts(
                colorbar=plot_params['style']['show_flagged_colorbar'],
                cmap=plot_params['style']['flagged_cmap'],
                responsive=True,
            ) * hv.QuadMesh([]).opts(
                colorbar=plot_params['style']['show_colorbar'],
                colorbar_position='left',
                cmap=plot_params['style']['unflagged_cmap'],
                responsive=True,
            )
        )
        return plot.opts(
            hv.opts.QuadMesh(tools=['hover'])
        )

    def _set_plot_colorbar(self, plot, plot_params, plot_type):
        ''' Update plot colorbar labels and limits
                plot_type is "unflagged" or "flagged"
        '''
        # Update colorbar labels if shown, else hide
        colorbar_key = plot_type + '_colorbar'
        if plot_params['plot'][colorbar_key]:
            c_label = plot_params['plot']['axis_labels']['c']['label']
            cbar_title = "Flagged " + c_label if plot_type == "flagged" else c_label

            plot = plot.opts(
                colorbar=True,
                backend_opts={
                    "colorbar.title": cbar_title,
                }
            )
        else:
            plot = plot.opts(
                colorbar=False,
            )

        # Update plot color limits
        color_limits = plot_params['plot']['color_limits']
        if color_limits:
            plot = plot.opts(
                clim=color_limits,
            )

        return plot

    def _update_color_range(self, plot_params):
        ''' Set the start/end range on the colorbar to min/max of plot data '''
        if self._gui_layout and 'data' in plot_params and 'data_range' in plot_params['data']:
            # Update range slider start and end to data min and max
            data_range = plot_params['data']['data_range']
            style_selectors = self._get_selector('style')
            range_slider = style_selectors[3][1]
            range_slider.start = data_range[0]
            range_slider.end = data_range[1]

    ###
    ### Update widget options based on MS
    ###
    def _get_selector(self, name):
        ''' Return selector group for name, for setting options '''
        selectors = self._gui_layout[2][1]
        selectors_index = {'file': 0, 'style': 1, 'sel': 2, 'axes': 3, 'agg': 4, 'iter': 5, 'title': 6}
        if name == "selectors":
            return selectors
        return selectors[selectors_index[name]]

    def _update_gui_axis_options(self):
        ''' Set gui options from ms data '''
        if 'data_dims' in self._ms_info:
            data_dims = self._ms_info['data_dims']
            axis_selectors = self._get_selector('axes')

            # Update options for x_axis selector
            x_axis_selector = axis_selectors.objects[0][0]
            x_axis_value = x_axis_selector.value
            x_axis_selector.options = data_dims
            if x_axis_value in data_dims:
                x_axis_selector.value = x_axis_value
            else:
                x_axis_selector.value = data_dims[1]

            # Update options for y_axis selector
            y_axis_selector = axis_selectors.objects[0][1]
            y_axis_value = y_axis_selector.value
            y_axis_selector.options = data_dims
            if y_axis_value in data_dims:
                y_axis_selector.value = y_axis_value
            else:
                y_axis_selector.value = data_dims[0]

            # Update options for vis_axis selector
            vis_axis_selector = axis_selectors.objects[2]
            vis_axis_value = vis_axis_selector.value
            if self._data.get_correlated_data('base') == 'SPECTRUM':
                vis_axis_selector.options = SPECTRUM_AXIS_OPTIONS
            else:
                vis_axis_selector.options = VIS_AXIS_OPTIONS
            if vis_axis_value in vis_axis_selector.options:
                vis_axis_selector.value = vis_axis_value
            else:
                vis_axis_selector.value = VIS_AXIS_OPTIONS[0]

            # Update options for selection selector
            selection_selectors = self._get_selector('sel')
            self._update_ps_selection_options(selection_selectors[0][0])
            self._update_ms_selection_options(selection_selectors[0][1])

            # Update options for agg axes selector
            agg_selectors = self._get_selector('agg')
            agg_selectors[1].options = data_dims

            # Update options for iteration axis selector
            iter_selectors = self._get_selector('iter')
            iter_axis_selector = iter_selectors[0][0]
            iter_axis_selector.options = ['None']
            iter_axis_selector.options.extend(data_dims)

    def _update_ps_selection_options(self, ps_selectors):
        ''' Set ProcessingSet gui options from ms summary '''
        if self._data and self._data.is_valid():
            summary = self._data.get_summary()
            if summary is None:
                return

            for selector in ps_selectors:
                if selector.name in PS_SELECTION_OPTIONS:
                    options = []
                    values = summary[PS_SELECTION_OPTIONS[selector.name]].values
                    for value in values:
                        if isinstance(value, list):
                            options.extend(value)
                        else:
                            options.append(value)
                    selector.options = sorted(list(set(options)))

    def _update_ms_selection_options(self, ms_selectors):
        ''' Set MeasurementSet gui options from ms data '''
        if self._data and self._data.is_valid():
            for selector in ms_selectors:
                selection_key = MS_SELECTION_OPTIONS[selector.name] if selector.name in MS_SELECTION_OPTIONS else None
                if selection_key:
                    if selection_key == 'data_group':
                        selector.options = list(super().data_groups())
                    else:
                        selector.options = super().get_dimension_values(selection_key)

    ###
    ### Callbacks for widgets which update other widgets
    ###
    def _set_filename(self, filename):
        ''' Set filename in text box from file selector value (list) '''
        if filename and self._gui_layout:
            file_selectors = self._get_selector('file')

            # Collapse FileSelector card
            file_selectors[1].collapsed = True

            # Change plot button color to indicate change unless ms path not set previously
            filename_input = file_selectors[0][0]
            if filename_input.value:
                self._update_plot_status(True)

            # Set filename from last file in file selector (triggers _update_plot())
            #filename_input.width = len(filename[-1])
            filename_input.value = filename[-1]


    def _set_iter_values(self, iter_axis):
        ''' Set up player with values when iter_axis is selected '''
        iter_axis = None if iter_axis == 'None' else iter_axis
        if iter_axis and self._gui_layout:
            iter_values = self._data.get_dimension_values(iter_axis)
            if iter_values:

                iter_selectors = self._get_selector('iter')

                # Update value selector with values and select first value
                iter_value_player = iter_selectors[1][0]
                if iter_axis == 'time':
                    if isinstance(iter_values[0], float):
                        iter_values = self._get_datetime_values(iter_values)
                    iter_value_player.format = get_time_formatter()
                elif iter_axis == 'frequency':
                    iter_value_player.format = NumeralTickFormatter(format='0,0.0000000')

                iter_value_player.options = iter_values
                iter_value_player.value = iter_values[0]
                iter_value_player.show_value = True

                # Update range inputs end values and select first
                iter_range_inputs = iter_selectors[1][1]
                last_iter_index = len(iter_values) - 1
                # range start
                iter_range_inputs[0][0].end = last_iter_index
                iter_range_inputs[0][0].value = 0
                # range end
                iter_range_inputs[0][1].end = last_iter_index
                iter_range_inputs[0][1].value = 0

    def _get_datetime_values(self, float_times):
        ''' Return list of float time values as list of datetime values for gui options '''
        time_attrs = self._data.get_dimension_attrs('time')
        datetime_values = []
        try:
            datetime_values = to_datetime(float_times, unit=time_attrs['units'], origin=time_attrs['format'])
        except TypeError:
            datetime_values = to_datetime(float_times, unit=time_attrs['units'][0], origin=time_attrs['format'])
        return list(datetime_values)

    def _update_plot_spinner(self, plot_clicked):
        ''' Callback to start spinner when Plot button clicked. '''
        if self._gui_layout:
            # Start spinner
            spinner = self._gui_layout[2][2][1]
            spinner.value = plot_clicked

    def _update_plot_status(self, inputs_changed):
        ''' Change button color when inputs change. '''
        if self._gui_layout:
            # Set button color
            button = self._gui_layout[2][2][0]
            button.button_style = 'solid' if inputs_changed else 'outline'

    def _update_plot_inputs(self):
        ''' Show inputs for raster plot in GUI '''
        if self._plot_params:
            inputs_column = self._gui_layout[0][1]
            inputs_column.clear()
            for param in self._plot_params:
                inputs_column.append(pn.pane.Str(param))

    ###
    ### Callbacks for widgets which update plot inputs
    ###
    def _set_title(self, title):
        ''' Set title from gui text input '''
        self._plot_inputs['title'] = title
        self._update_plot_status(True) # Change plot button to solid

    def _set_style_params(self, unflagged_cmap, flagged_cmap, show_colorbar, show_flagged_colorbar):
        self.set_style_params(unflagged_cmap, flagged_cmap, show_colorbar, show_flagged_colorbar)
        self._update_plot_status(True) # Change plot button to solid

    def _set_color_range(self, color_mode, color_range):
        ''' Set style params from gui '''
        color_mode = color_mode.split()[0]
        color_mode = None if color_mode == 'No' else color_mode
        self._plot_inputs['color_mode'] = color_mode
        self._plot_inputs['color_range'] = color_range
        self._update_plot_status(True) # Change plot button to solid

    def _set_axes(self, x_axis, y_axis, vis_axis):
        ''' Set plot axis params from gui '''
        self._plot_inputs['x_axis'] = x_axis
        self._plot_inputs['y_axis'] = y_axis
        self._plot_inputs['vis_axis'] = vis_axis
        self._update_plot_status(True) # Change plot button to solid

    def _set_aggregation(self, aggregator, agg_axes):
        ''' Set aggregation params from gui '''
        aggregator = None if aggregator== 'None' else aggregator
        self._plot_inputs['aggregator'] = aggregator
        self._plot_inputs['agg_axis'] = agg_axes # ignored if aggregator not set
        self._update_plot_status(True) # Change plot button to solid

# pylint: disable=too-many-arguments, too-many-positional-arguments
    def _set_iteration(self, iter_axis, iter_value_type, iter_value, iter_start, iter_end, subplot_rows, subplot_columns):
        ''' Set iteration params from gui '''
        iter_axis = None if iter_axis == 'None' else iter_axis
        self._plot_inputs['iter_axis'] = iter_axis
        self._plot_inputs['subplots'] = (subplot_rows, subplot_columns)

        if iter_axis:
            if iter_value_type == 'By Value':
                # Use index of iter_value for tuple
                if self._data and self._data.is_valid():
                    iter_values = self._data.get_dimension_values(iter_axis)
                    iter_index = iter_values.index(iter_value)
                    self._plot_inputs['iter_range'] = (iter_index, iter_index)
            else:
                # 'By Range': use range start and end values for tuple
                self._plot_inputs['iter_range'] = (iter_start, iter_end)
        else:
            self._plot_inputs['iter_range'] = None
        self._update_plot_status(True) # Change plot button to solid
# pylint: enable=too-many-arguments, too-many-positional-arguments

# pylint: disable=too-many-arguments, too-many-positional-arguments, unused-argument
    def _set_ps_selection(self, query, name, intents, scan_name, spw_name, field_name, source_name, line_name):
        ''' Select ProcessingSet from gui using summary columns '''
        inputs = locals()
        ps_selection = {}
        for key, val in inputs.items():
            if key in PS_SELECTION_OPTIONS.values() and val:
                ps_selection[key] = val
        self._plot_inputs['ps_selection'] = ps_selection
        self._update_plot_status(True) # Change plot button to solid

    def _set_ms_selection(self, data_group, datetime, baseline, antenna1, antenna2, frequency, polarization):
        ''' Select MeasurementSet from gui using data group, dimensions, and coordinates '''
        inputs = locals()
        inputs['time'] = inputs.pop('datetime')
        ms_selection = {}
        for key, val in inputs.items():
            if key in MS_SELECTION_OPTIONS.values() and val:
                ms_selection[key] = val
        self._plot_inputs['ms_selection'] = ms_selection
        self._update_plot_status(True) # Change plot button to solid
# pylint: enable=too-many-arguments, too-many-positional-arguments, unused-argument
